#!/usr/bin/perl
#
# SCoRE - Search a Corpus for Regular Expressions
#
# Perl module for BEE-specific routines
#
# (c) Matthew Purver, King's College London, 2002
#
# $Log: $
#

package BEE;

# export subroutines
require Exporter;
our @ISA = qw( Exporter );
our @EXPORT = qw( checkHeaderType 
		  getHeaderInfo
		  getSentence
		  removeSentence
		  noTags
		  noTagsAtAll
		  highlight
		  fullPath
		  skipHeader 
		  skipAlign 
		  stringToRegExp );
our @EXPORT_OK = qw( %includeTypes $defaultIncludeType %excludeTypes $defaultExcludeType @dataDirs
		     $corpusFilePath $corpusListPath );

# imports
use SCoRE qw( $corpusPath $startMark $endMark $startTag $endTag $sentNum $useTurns $startDir $showComments );

# defines
$corpusFilePath = $corpusPath;			# path to corpus files
$corpusListPath = $corpusPath;			# path to corpus file lists
$startDir = ( $startDir eq '' ) ? '*/*/*' : '$startDir/*'; # modify default dir spec

$turnSep = '<tr><td.*?\/td>\s*<td>\s*<b>[^<>]*<\/b>';
$sentSep = '(?:(?<=[.?!])\s+(?!<|$))|(?:<tr><td.*?\/td>\s*<td>\s*<b>[^<>]*<\/b>)';
$uttSep = $useTurns ? $turnSep : $sentSep;	# sentence/turn separator
$turnSepC = '<tr><td.*?\/td>\s*<td>\s*<b>([^<>]*)<\/b>';
$sentSepC = '(?:(?<=[.?!])\s+(?!<|$))|(?:<tr><td.*?\/td>\s*<td>\s*<b>([^<>]*)<\/b>)';


# discard to end of header section (or fail)
sub skipHeader {
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*($uttSep)/$1/ );
    }
    return 0;
}


# dummy function (no align section)
sub skipAlign {
    return 1;
}


# initialize & read to end of header section (or fail)
sub getHeader {
    $lineBuffer = '';
    while ( <STDIN> ) {
	chomp( $lineBuffer .= $_ );
	return 1 if ( $lineBuffer =~ /$uttSep/ );
    }
    return 0;
}


# pull <title> and <h3> information from header
sub getHeaderDescription {
    my ( $info, $title ) = ( 'none', 'none' );
    ( $lineBuffer =~ /<title>([^<>]*)<\/title>/i ) and ( $info = $1 );
    ( $lineBuffer =~ /<h3>([^<>]*)<\/h3>/i ) and ( $title = $1 );
    # return info when header finished
    return ( $info, $title );
}

# no type info, so just get file description
sub checkHeaderType {
    # get file description
    getHeader() or return ();
    my ( $info, $title ) = getHeaderDescription();

    # discard header from buffer and return info
    $lineBuffer =~ s/^.*?($uttSep)/$1/;
    return ( $info, $title );
}


# set up CP names & corresponding display colours (don't need header info though)
sub getHeaderInfo {

    @colours = ( 'Gray', 'Red', 'Purple', 'Maroon', 'Fuchsia', 
  		 'Green', 'Lime', 'Olive', 'Yellow', 'Navy', 'Blue', 'Teal', 'Aqua' );

    # get file description
    getHeader() or return ();
    my ( $info, $title ) = getHeaderDescription();

    # set up names & colours for global IDs
    $people{ 'Student:' } = 'Student';
    $people{ 'Tutor:' } = 'Tutor';
    $whoColours{ 'Student:' } = 'Blue';
    $whoColours{ 'Tutor:' } = 'Green';
    my @cpinfo = ( 'Student', 'Tutor' );

    # discard header from buffer and return info
    $lineBuffer =~ s/^.*?($uttSep)/$1/;
    return ( $info, $title, @cpinfo );
}


sub noTagsAtAll {
    my @string = @_;

    for ( @string ) {
	s/\{[^}]*?\}//g;
	s/\[([^\[\]]*)\]/$1/g;
	s/$turnSep/| /go;
	s/$sentSep/ /go;
        s/<[^<>]*>//g;
    }

    return @string;
}


# process a sentence SGML -> HTML suitable for display
$odd = 1;
sub noTags {
    my ( $notags ) = @_;
    my $l = '<em>&lt;';   # MUST contain > chars
    my $r = '&gt;</em>';  # MUST contain > chars
    my $lastNum = 1;

    for ( $notags ) {

	# make {} tags visible
	s/\{([^}]*?)\}/$l$1$r/go;

	# strikeout truncated words
#	s/(\w+)-(\s)/<s>$1<\/s>$2/go;

	# colour-code overlapping speech
	$odd and s/\[([^\[\]]*)\]/"<font COLOR=" . $colours[++$iOverlapColour % @colours] . ">$1<\/font>"/ge and ( $odd = 0 );
	!$odd and s/\[([^\[\]]*)\]/"<font COLOR=" . $colours[$iOverlapColour % @colours] . ">$1<\/font>"/ge and ( $odd = 1 );

	# show comments
	s/($l)comment value=\"(.*?)\"$r/$showComments ? " <em>[$2]<\/em>" : ''/ge;

	# table formatting for utterances/turns
        s/<\/td><\/tr>\s*//g; # remove existing HTML table row ends
	s/$turnSepC/"<\/td><\/tr>\n<tr><td bgcolor=$whoColours{$1}><em>$people{$1}<\/em><\/td><td>" . ( $sentNum ? "<font size=-2>x<\/font>" : '' ) . "<\/td><td>"/goe;
	s/$sentSepC/"<\/td><\/tr>\n<tr><td><\/td><td>" . ( $sentNum ? "<font size=-2>x<\/font>" : '' ) . "<\/td><td>"/goe;
	s/^\s*(?!<)/"<\/td><\/tr>\n<tr><td><\/td><td>" . ( $sentNum ? "<font size=-2>x<\/font>" : '' ) . "<\/td><td>"/goe;

    }

    # finally, replace __B/E__ markers
    $notags =~ s/$startMark/$startTag/go;
    $notags =~ s/$endMark/$endTag/go;

    return $notags;
}


# convert search string to reg-exp suitable for searching SGML
sub stringToRegExp {
    my ( $arg, $caseSens, $punctSens, $strictPoS ) = @_;

    $pu = '[,.?!;:]';
    $sp = ( $punctSens ? '\s*' : "$pu?\\s*" );
    $cs = ( $caseSens ? '' : '(?i)' );
    $br = $uttSep;
    $wd = '[\'\w]'; # words may contain ' chars (e.g. n't)
    $mk = '';   # optional marker chars
    my $nSent = 0;

    # translate wildcards to regexp
    for ( $arg ) {

	# first count the number of sentences/turns
	$nSent += ( @tmp = /(^|\s)\|/g );

	# replace whitespace by tabs to make it easier to replace uniquely later
	s/\s+/\t/go;

	# convert wildcards to regexp equivalents
	s/\(/\(\?:/g;         # ( -> (?:  (to prevent grouping also gathering)
	s/(?<!\])\*/$wd\*/go;        # * -> ['\w]*
	s/(?<!\])\+/$wd\+/go;        # + -> ['\w]+
	s/([^\s\(\)])\?/$1$wd/go;    # ? -> ['\w]  (unless preceded by whitespace/bracket)
	s/([^\(\)])\?(\S)/$1$wd$2/go;    # ? -> ['\w]  (unless preceded by bracket/followed by whitespace)
	s/(\S+),/$1\t\.\*\?/go;  # , -> .*? (if not preceded by whitespace)

        # spot PoS tag searches
        /</ and print "Sorry - no PoS tagging in BEE" and die "PoS search attempted";

	# deal with sentence/turn breaks (don't allow .* if ^, $ adjacent)
        s/(\$\t)(\|)(\t\^)/$1$br$3/go;
        s/(^|\t)(\|)(\t\^)/$1.*?$br$3/go;
        s/(\$\t)(\|)(\t|$)/$1$br.*?$3/go;
        s/(^|\t)(\|)(\t|$)/$1.*?$br.*?$3/go;

	# add required tags to ^, $, keeping (^/$) only if actually at beginning/end
	s/(^)(\^)(\t|$)/($2)\\s*(?:<[^<>]*>\\s*)*$3/go;
	s/(\t)(\^)(\t|$)/$1\\s*(?:<[^<>]*>\\s*)*$3/go;
        s/(^|\t)(\$)($)/$1$pu?\\s*(?:<[^<>]*>\\s*)*($2)$3/go;
        s/(^|\t)(\$)(\t)/$1$pu?\\s*(?:<[^<>]*>\\s*)*$3/go;

	# mark punctuation marks to prevent \b
	s/(^|\t)($pu)(\t|$)/$1<$2>$3/go;

        # add boundaries to words & capture (rearranging optional brackets)
  	s/(^|\t)([^<\t.]+)/$1($cs$mk\\b$mk$2$mk\\b$mk)/go; # no \. so don't get old commas
  	s/(^|\t)(\.\*\?[^<\t.]+)/$1($2)/go; # take care of |s which now contain \.
        s/(\\b$mk)(\(\?:)([^<\t.]+)(\)\?)($mk\\b)/$2$1$3$5$4/go;

	# escape punctuation marks & capture
	s/<($pu)>/(\\$1)/go;

	# and replace tabs with required separator
	s/(^|\t)(\.\*\?)(\t|$)/$2/go;
	s/\t/$sp/go;

	# now put brackets round the whole thing for capturing (UNLESS empty)
	# (remembering to increment all capture-counters ...)
	s/^/(/;
	s/$/)/;
	s/^\(\)$//;
	s/\\(\d+)/'\\' . ($1+1)/ge;

    }
    return ( $arg, $nSent );

}


sub getSentence {

    while ( <STDIN> ) {
  	chomp( $lineBuffer .= $_ );

    	# beginning of next sentence? return complete one
  	if ( $lineBuffer =~ /($uttSep).*?($uttSep)/o ) {
  	    $lineBuffer =~ s/^(.*?($uttSep).*?)($uttSep)/$3/o;
  	    return $1;
  	}
    }
    
    # end of file? return any dregs
    $lineBuffer =~ s/^(.*)$//o;
    return $1;

}


sub removeSentence {
    my( $string ) = @_;

    # remove first sentence (or whole thing) from buffer string
    ( $string =~ s/(^.*?$uttSep.+?)($uttSep)/$2/o ) or ( $string =~ s/(.*)//o );

    # return what's left and what was removed )
    return ( $string, $1 );
}


sub highlight {
    my( $text ) = @_;

    for ( $text ) {
	s/^/$startMark/o;
	s/$/$endMark/o;
	s/($sentSep)/$1$startMark/go;
    }
    return $text;
}


sub fullPath {
    my( $file ) = @_;

    ( $file =~ /$corpusFilePath/ ) or ( $file = "$corpusFilePath/$file" );
    return $file;
}


%includeTypes = ();
$defaultIncludeType = '';
%excludeTypes = ();
$defaultExcludeType = '';
@dataDirs = ( 'PilotStudy/Logs', 'FinalExperiment/Logs' );

# ensure package returns 1 to keep Perl happy
1;
