#!/usr/bin/perl
#
# SCoRE - Search a Corpus for Regular Expressions
#
# Perl module for BNC-specific routines
#
# (c) Matthew Purver, King's College London, 2001
#
# $Log: BNC.pm,v $
# Revision 1.1  2001/11/07 14:47:18  purver
# Initial revision
#

package BNC;

# export subroutines
require Exporter;
our @ISA = qw( Exporter );
our @EXPORT = qw( checkHeaderType 
		  getHeaderInfo
		  getSentence
		  removeSentence
		  noTags 
		  noTagsAtAll
		  highlight
		  fullPath
		  skipHeader 
		  skipAlign 
		  stringToRegExp );
our @EXPORT_OK = qw( %includeTypes $defaultIncludeType %excludeTypes $defaultExcludeType @dataDirs
		     $corpusFilePath $corpusListPath );

# imports
use SCoRE qw( $corpusPath $startMark $endMark $questionMark $startTag $endTag $sentNum $useTurns $startDir $showComments );

# defines
$corpusListPath = $corpusPath;		# path to corpus file lists
$corpusFilePath = "$corpusPath/Texts";	# path to corpus files

$turnSep = '<[up]\b[^<>]*>\s*'; # turn separator (or paragraph for written part)
$sentSep = '(?:<[up]\b[^<>]*>\s*)?<s\b[^<>]*>\s*'; # sentence separator
$uttSep = $useTurns ? $turnSep : $sentSep; # sentence/turn separator

if ( $startDir eq '' ) {	# modify default dir spec
    $startDir = '*/*/*';
}
else {
    # otherwise adjust to full length if necessary
    if ( $startDir =~ /(\w)(\/\w\w)?(\/\w\w\w)?/ ) {
	$startDir = $1 . ( defined( $2 ) ? $2 : '/*' ) . ( defined( $3 ) ? $3 : '/*' );
    }
}


# discard to end of <teiHeader> section (or fail)
sub skipHeader {
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*\/teiHeader>// );
    }
    return 0;
}


# discard to end of <align> section (or fail)
sub skipAlign {
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*\/align>// );
    }
    return 0;
}


# read to end of <teiHeader> section (or fail)
sub getHeader {
    $lineBuffer = '';
    while ( $lineBuffer .= <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ /\/teiHeader>/ );
    }
    return 0;
}



# get file description from header
sub getHeaderDescription {
    my $maxDesc = 3;
    my $info = ' ';
    my $title = 'none';

    # get interesting information if present
    ( $lineBuffer =~ /<titleStmt.*<title[^<>]*>\s*([^<>]*?)\s*<\/title.*<\/titleStmt/ ) and ( $title  = $1 );
    if ( $lineBuffer =~ /catRef\s+target=\"[^\"]*wridom/ ) {
	my @title = ( $lineBuffer =~ /<biblStruct.*<title[^<>]*>\s*([^<>]*?)\s*<\/title/g );
	my @publisher = ( $lineBuffer =~ /<publisher[^<>]*>\s*([^<>]*?)\s*<\/publisher/g );
	$info = "@title - @publisher";
    }
    else {
	my @activity = ( $lineBuffer =~ /<activity[^<>]*>\s*([^<>]*?)\s*<\/activity/g );
	my @locale = ( $lineBuffer =~ /<locale[^<>]*>\s*([^<>]*?)\s*<\/locale/g );
	my $max = ( @activity > @locale ) ? @activity : @locale;
	$max = ( $max > $maxDesc ) ? $maxDesc : $max;
	for ( my $iDesc = 0; $iDesc < $max; $iDesc++ ) {
	    $info .= ( defined( $locale[$iDesc] ) ? $locale[$iDesc] : '?' );
	    $info .= ' - ' . ( defined( $activity[$iDesc] ) ? $activity[$iDesc] : '?' ) . '; ';
	}
    }

    # and return it
    return ( ( $info or 'none' ), $title );
}

# check <catRef target=> types suitable (or fail), and return <locale> & <activity> info
# if available
sub checkHeaderType {
    my ( $arg, @inc, @exc, $info, $title, $iDesc );
    my @args = @_;
    my $success = 0;

    # build list of include/exclude values
    foreach $arg (@args) {
	push( @inc, $arg ) if ( $arg =~ s/^\+// );
	push( @exc, $arg ) if ( $arg =~ s/^-// );
    }

    # get whole <teiHeader> section
    getHeader() or return ();

    # get <catRef> target attribute
    if ( $lineBuffer =~ /<catRef target=\"([^\"]*)\"/ ) {
	# split into array of values
	@chk = split( /\s+/, $1 );
	foreach $chk (@chk) {
	    # fail if any exclude values match
	    foreach $exc (@exc) {
		( $chk =~ /$exc/ ) and return ();
	    }
	    # succeed if only one include value matches
	    unless ( $success ) {
		foreach $inc (@inc) {
		    ( $chk =~ /$inc/ ) and ( $success = 1 ) and last;
		}
	    }
	}
    }
    return () unless $success;

    # get file description
    ( $info, $title ) = getHeaderDescription();
    
    # discard <teiHeader> section from buffer and return
    $lineBuffer =~ s/^.*\/teiHeader>//;
    return ( $info, $title );
}


# set up CP names & corresponding display colours
sub getHeaderInfo {
    my ( $iAnon, $iColour, $key, $value, $info, $title, @cpinfo );

    @colours = ( 'Gray', 'Maroon', 'Red', 'Purple', 'Fuchsia', 
		 'Green', 'Lime', 'Olive', 'Yellow', 'Navy', 'Blue', 'Teal', 'Aqua' );

    # get whole <teiHeader> section and file description
    getHeader() or return ();
    ( $info, $title ) = getHeaderDescription();

    # get CP names from <person> info if present, distinguishing between 'None' entries
    $iAnon = 0;
    $iColour = 0;
    $iCP = -1;
    %people = ( $lineBuffer =~ /<person[^<>]*?id=\"([^\"]*)\"[^<>]*>(.*?)<\/person/g );
    while (($key, $value) = each %people) {
	if ( ( $value =~ /<name>\s*([^<>]+?)\s*<\/name/ ) and ( $1 ne 'None' ) ) {
	    $people{ $key } = $1;
	}
	else {
	    $people{ $key } = "Anon " . ++$iAnon;
	}
	$cpinfo[++$iCP] = $people{$key};
	( $value =~ /<age>\s*([^<>]+?)\s*<\/age/ ) and ( $cpinfo[$iCP] .= " ($1)" );
	( $value =~ /<occupation>\s*([^<>]+?)\s*<\/occupation/ ) and ( $cpinfo[$iCP] .= ", $1" );
	( $value =~ /<dialect>\s*([^<>]+?)\s*<\/dialect/ ) and ( $cpinfo[$iCP] .= ", $1" );
	$whoColours{ $key } = $colours[($iColour++) % @colours];
    }

    # set up names & colours for global IDs
    $people{ 'PS000' } = 'Unknown';
    $people{ 'PS001' } = 'Unknowns';
    $whoColours{ 'PS000' } = $whoColours{ 'PS001' } = 'White';

    # discard <teiHeader> section from buffer and return info
    $lineBuffer =~ s/^.*\/teiHeader>//;
    return ( $info, $title, @cpinfo );
}


sub noTagsAtAll {
    my @string = @_;

    for ( @string ) {
	s/\&mdash\;/\&\#8212\;/g;
	s/\&[be]quo\;/\&quot\;/g;
	s/$turnSep/| /go;
	s/$sentSep/ /go;
	s/<[^<>]*>//g;
    }

    return @string;
}


# process a sentence SGML -> HTML suitable for display
sub noTags {
    my ( $notags ) = @_;
    my $l = '<em>&lt;';   # MUST contain > chars
    my $r = '&gt;</em>';  # MUST contain > chars
    my $odd = 1;

    ( $ptrColour = 0 ) unless defined $ptrColour;

    for ( $notags ) {

	# convert SGML entities
	s/\&mdash\;/\&\#8212\;/g;
	s/\&[be]quo\;/\&quot\;/g;

	# remove tags we don't want
	s/<w\s+\w{3}(-\w{3})?>//go;
	s/<c\s+PU\w>//go;
	s/<\/[up]>//go;
	s/<\/div(\d)?>//go;
	s/<(\/)?s?text[^>]*>//go;
	s/<(\/)?body[^>]*>//go;
	s/<bncDoc[^>]*>//go;
	s/<(\/)?align>//go;
	s/<(\/)?(pb|lb)[^>]*>//go;
	s/<loc[^>]*>//go;

	# make those we do want visible
	s/<([^>]*?)>/$l$1$r/go;

	# replace <trunc> sections by strikeout
	s/($l)trunc$r/<s>/go;
	s/($l)\/trunc$r/<\/s>/go;

	# make <sic> sections easier to read
	s/($l)sic[^>]*$r//go;
	s/($l)\/sic$r/<em>\[sic\]<\/em>/go;

	# make <corr> sections easier to read
#	s/($l)corr[^>]*sic=\"([^\"]*)\"[^>]*$r/<em>\[corrected from $2\]<\/em>/go;
	s/($l)corr[^>]*sic=(\"|\')([^\"]*)\2[^>]*$r/$1corrected from $3$r:[/go;
	s/($l)\/corr$r/]/go;

	# make <corr> sections easier to read
	s/($l)quote[^>]*$r/<em>\[begin quote\]<\/em>/go;
	s/($l)\/quote$r/<em>\[end quote\]<\/em>/go;

	# make anonymisation gaps easier to read
	s/($l)gap\s[^>]*?desc=(\"|\')([^>]*?)\2.*?$r/<em>\[$3\]<\/em>/go;
	s/($l)gap\s[^>]*?desc=(\w*)[^>]*?$r/<em>\[$2\]<\/em>/go;

	# and same for events
	s/($l)event\s[^>]*?desc=(\")?([^>\"]*)(\")?/$1$3/go;
	s/($l)vocal\s[^>]*?desc=(\")?([^>\"]*)(\")?/$1$3/go;

	# and same for shifts
	s/($l)shift\s[^>]*?new=(\")?([^>\"]*)(\")?$r/$1$3$r:[/go;
	s/($l)shift$r/]/go;

	# insert name for unclear utterances
	if ( /($l)unclear\s[^>]*?who=(\w+)/o ) {
	    $name = ( defined( $people{$2} ) ? $people{$2} : $2 );
	    s/(($l)unclear\s[^>]*?)who=(\w+)/$1 \[$name\]/go;
	}

	# replace <trunc> sections by strikeout
	s/($l)hi\s+[^>]*$r/<em>/go;
	s/($l)\/hi$r/<\/em>/go;

	# replace <div>, <bncDoc> boundaries
	s/($l)div\s+[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Conversation -----------------/go;
	s/($l)div1[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Chapter -----------------/go;
	s/($l)div2[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Section -----------------/go;
	s/($l)div3[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Sub-section -----------------/go;
	s/($l)div4[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Sub-sub-section -----------------/go;
	s/($l)\/bncDoc($r)/<\/td><\/tr><tr><td colspan=3>----------------- End -----------------/o;
	s/($l)\/bncDoc($r)//go;

    }

    # colour-coding for overlapping <ptr> sections
    while ( $notags =~ /ptr\s+target=(\w+)$r/o ) {
	unless ( defined $ptrColourTags{$1} ) {
	    if ( $odd ) {
		$ptrColourTags{$1} = "<font COLOR=$colours[($ptrColour++) % @colours]>";
	    }
	    else {
		$ptrColourTags{$1} = '</font>';
	    }
	    $odd = !$odd;
	}
	$notags =~ s/($l)ptr\s+target=(\w+)$r/$ptrColourTags{$2}/o;
    }

    # table formatting for u/s tags
    for ( $notags ) {

	s/($l)u\s+who=(\w+)\s*$r\s*(?=($l)s\s)/<\/td><\/tr><tr><td nowrap bgcolor=$whoColours{$2}><em>$people{$2}<\/em><\/td><td>/go; # if <s> coming next, act normal
	s/($l)u\s+who=(\w+)\s*$r\s*/"<\/td><\/tr><tr><td nowrap bgcolor=$whoColours{$2}><em>$people{$2}<\/em><\/td>" . ( $sentNum ? "<td><\/td>" : "" ) . "<td>"/goe; # otherwise insert box for (absent) sent num

	s/($l)p(\s+[^>]*)?$r\s*(?=($l)s\s)/<\/td><\/tr><tr><td nowrap><em>-<\/em><\/td><td>/go; # if <s> coming next, act normal
	s/($l)p(\s+[^>]*)?$r\s*/"<\/td><\/tr><tr><td nowrap><em>-<\/em><\/td>" . ( $sentNum ? "<td><\/td>" : "" ) . "<td>"/goe; # otherwise insert box for (absent) sent num

	s/(?<=<td>)($l)s\s+n=\"(\d+)\"\s*$r/( $sentNum ? "<font SIZE=-2>$2<\/font><\/td><td>" : "" )/goe; # if preceded by <u>, just new cell
	s/(($l)(head|caption|note|label|item)[^>]*($r))\s*($l)s\s+n=\"(\d+)\"\s*$r/"<\/td><\/tr><tr><td><\/td>" . ( $sentNum ? "<td><font SIZE=-2>$6<\/font><\/td>" : "" ) . "<td>$1"/goe; # otherwise start new line too
	s/($l)s\s+n=\"(\d+)\"\s*$r/"<\/td><\/tr><tr><td><\/td>" . ( $sentNum ? "<td><font SIZE=-2>$2<\/font><\/td>" : "" ) . "<td>"/goe; # otherwise start new line too

    }

    for ( $notags ) {

	# headers etc.
	s/($l)head[^>]*main[^>]*($r)/<h3><u>/goi;
	s/($l)head[^>]*sub[^>]*($r)/<h3>/goi;
	s/($l)(head|caption)[^>]*($r)/<h3><u>/go;
	s/($l)\/(head|caption)($r)/<\/u><\/h3>/go;

	s/($l)(note|label|item)[^>]*($r)/&nbsp;&nbsp;&nbsp;/go;
	s/($l)\/(note|label|item)($r)//go;

	s/($l)[^>]*list[^>]*($r)//go;

    }

    # finally, replace __B/E__ markers
    $notags =~ s/$startMark/$startTag/go;
    $notags =~ s/$endMark/$endTag/go;

    return $notags;
}

# convert search string to reg-exp suitable for searching SGML
sub stringToRegExp {
    my ( $arg, $caseSens, $punctSens, $strictPoS ) = @_;

    $sp = ( $punctSens ? '\s*' : '(?:<c PU\w>\W)?\s*' );
    $cs = ( $caseSens ? '' : '(?i)' );
    $br = $uttSep;
    $wd = '[\'\w]'; # words may contain ' chars (e.g. n't)
    my $nSent = 0;

    # translate wildcards to regexp
    for ( $arg ) {

	# first count the number of sentences/turns
	$nSent += ( @tmp = /(^|\s)\|/g );

#  	# ensure 'd, 're, n't etc. are separate words
	s/(n\'t)/ $1/g;
  	s/(?<!n)\'(?!t)/ \'/g;

	# replace whitespace by tabs to make it easier to replace uniquely later
	s/\s+/\t/g;

	# convert wildcards to regexp equivalents
	s/\(/\(\?:/g;         # ( -> (?:  (to prevent grouping also gathering)
	s/(?<!\])\*/$wd\*/go;        # * -> ['\w]*
	s/(?<!\])\+/$wd\+/go;        # + -> ['\w]+
	# save ?s which are *not* DOS-style wildcards
	s/([\(\)])\?/$1$questionMark/go;    # if preceded by bracket
	s/(^|\t)\?((,)?\t|$)/$1$questionMark$2/go;    # if surrounded by whitespace (maybe with comma)
	s/(^|\t)\?((,)?\t|$)/$1$questionMark$2/go;    # if surrounded by whitespace (maybe with comma)
	s/(\[[^\]]*)\?([^\]]*\])/$1$questionMark$2/go;    # if in [] character class
	# convert ?s left
	s/\?/$wd/go;    # ? -> ['\w]  (unless preceded by whitespace/bracket)
#	s/([^\s\(\)])\?/$1$wd/go;    # ? -> ['\w]  (unless preceded by whitespace/bracket)
#	s/([^\(\)])\?(\S)/$1$wd$2/go;    # ? -> ['\w]  (unless preceded by bracket/followed by whitespace)
	# unsave saved ?s
	s/$questionMark/\?/go;
	s/(\S+),/$1\t\.\*\?/go;  # , -> .*? (if not preceded by whitespace)

	# first deal with PoSs (currently the only <> tags in $arg)
#	s/<([^<>]+)>/<w (?:\\w{3}-)?$1(?:-\\w{3})?>(\\w+)\\b/go;
#	s/<([^<>]+)>/<w (?:\\w{3}-)?$1(?:-\\w{3})?>($wd+)\\b/go;
	s/<([^<>]+)>(\S*)/"<w " . ( $strictPoS ? '' : "(?:\\w{3}-)?" ) . "$1(?:-\\w{3})?>(" . 
	    ( ( $2 eq '' ) ? "$wd+" : "$cs$2" ) . ")\\b"/goe;

	# deal with numbered sentence breaks
        s/(\$\t)\|(\S+)(\t\^)/$1(<s\\s+n=\\"$2\\">)$3/go;
        s/(^|\t)\|(\S+)(\t\^)/$1(.*?<s\\s+n=\\"$2\\">)$3/go;
        s/(\$\t)\|(\S+)(\t|$)/$1(<s\\s+n=\\"$2\\">.*?)$3/go;
        s/(^|\t)\|(\S+)(\t|$)/$1(.*?<s\\s+n=\\"$2\\">.*?)$3/go;

	# deal with sentence/turn breaks (don't allow .* if ^, $ adjacent)
        s/(\$\t)(\|)(\t\^)/$1($br)$3/go;
        s/(^|\t)(\|)(\t\^)/$1(.*?$br)$3/go;
        s/(\$\t)(\|)(\t|$)/$1($br.*?)$3/go;
        s/(^|\t)(\|)(\t|$)/$1(.*?$br.*?)$3/go;

	# add required tags to ^, $, keeping (^/$) only if actually at beginning/end
	s/(^)(\^)(\S+)(\t|$)/($2)\\s*(?:<[^<>]*>\\s*)*<s\\s+n=\\"$3\\">(?:<[^<>]*>\\s*)*$4/go;
	s/(\t)(\^)(\S+)(\t|$)/$1\\s*(?:<[^<>]*>\\s*)*<s\\s+n=\\"$3\\">(?:<[^<>]*>\\s*)*$4/go;
	s/(^)(\^)(\t|$)/($2)\\s*(?:<[^<>]*>\\s*)*$3/go;
	s/(\t)(\^)(\t|$)/$1\\s*(?:<[^<>]*>\\s*)*$3/go;
        s/(^|\t)(\$)($)/$1(?:<c PU\\w>\\W)?\\s*(?:<[^<>]*>\\s*)*($2)$3/go;
        s/(^|\t)(\$)(\t)/$1(?:<c PU\\w>\\W)?\\s*(?:<[^<>]*>\\s*)*$3/go;

	# add PoS tag to punctuation marks
	s/(^|\t)([,.?!;:()])(\t|$)/$1<c PU\\w>(\\$2)$3/go;

        # add PoS tag to words, inside optional brackets if necessary
  	s/(^|\t)(\(\?:)([^.<\t]+)(\)\?)(\t|$)/$1$2<w \\w{3}(?:-\\w{3})?>($cs$3)\\b$4$5/go;
  	s/(^|\t)(\(\?:)([^.<\t]+)(\)\?)(\t|$)/$1$2<w \\w{3}(?:-\\w{3})?>($cs$3)\\b$4$5/go;
  	s/(^|\t)([^.<\t]+)(\t|$)/$1<w \\w{3}(?:-\\w{3})?>($cs$2)\\b$3/go;
  	s/(^|\t)([^.<\t]+)(\t|$)/$1<w \\w{3}(?:-\\w{3})?>($cs$2)\\b$3/go;

	# and replace tabs with required separator
	s/(^|\t)(\.\*\?)(\t|$)/$2/go;
	s/\t/$sp/go;

	# now put brackets round the whole thing for capturing (UNLESS empty)
	# (remembering to increment all capture-counters ...)
	s/^/(/;
	s/$/)/;
	s/^\(\)$//;
	s/\\(\d+)/'\\' . ($1+1)/ge;

    }
    return ( $arg, $nSent );

}


sub getSentence {

    while ( <STDIN> ) {
  	chomp( $lineBuffer .= $_ );

    	# beginning of next sentence? return complete one
  	if ( $lineBuffer =~ /$uttSep.*?$uttSep/o ) {
  	    $lineBuffer =~ s/^(.*?$uttSep.*?)($uttSep)/$2/o;
  	    return $1;
  	}
    }
    
    # end of file? return any dregs
    $lineBuffer =~ s/^(.*)$//o;
    return $1;

}


sub removeSentence {
    my( $string ) = @_;

    # remove first sentence (or whole thing) from buffer string
    ( $string =~ s/(^.*?$uttSep.+?)($uttSep)/$2/o ) or ( $string =~ s/(.*)//o );

    # return what's left and what was removed )
    return ( $string, $1 );
}


sub highlight {
    my( $text ) = @_;

    for ( $text ) {
	s/^/$startMark/o;
	s/$/$endMark/o;
	s/($turnSep($sentSep)?)/$1$startMark/go;
	s/($sentSep)(?!$startMark)/$1$startMark/go;
    }
    return $text;
}


sub fullPath {
    my( $file ) = @_;

    $file =~ s/^((([A-Z0-9])[A-Z0-9])[A-Z0-9])$/$3\/$2\/$1/;
    ( $file =~ /$corpusFilePath/ ) or ( $file = "$corpusFilePath/$file" );

    return $file;
}


%includeTypes = (
		 'wridom' => 'Written',
		 'spolog1' => 'Spoken (monologue)',
		 'spolog2' => 'Spoken (dialogue)'
		 );
$defaultIncludeType = 'spolog2';

%excludeTypes = (
		 'wridom1' => 'Written: Imaginative',
		 'wridom2' => 'Written: Informative (natural &amp; pure science)', 
		 'wridom3' => 'Written: Informative (applied science)',
		 'wridom4' => 'Written: Informative (social science)',
		 'wridom5' => 'Written: Informative (world affairs)',
		 'wridom6' => 'Written: Informative (commerce &amp; finance)',
		 'wridom7' => 'Written: Informative (arts)',
		 'wridom8' => 'Written: Informative (belief &amp; thought)',
		 'wridom9' => 'Written: Informative (leisure)',
		 'scgdom1' => 'Spoken: Context-governed (educational)',
		 'scgdom2' => 'Spoken: Context-governed (business)',
		 'scgdom3' => 'Spoken: Context-governed (public)',
		 'scgdom4' => 'Spoken: Context-governed (leisure)'
		 );
$defaultExcludeType = '';

@dataDirs = (

	     'A', 'A/A0', 'A/A1', 'A/A2', 'A/A3', 'A/A4', 'A/A5', 'A/A6', 'A/A7', 'A/A8', 'A/A9',
	     'A/AA', 'A/AB', 'A/AC', 'A/AD', 'A/AE', 'A/AH', 'A/AJ', 'A/AK', 'A/AL', 'A/AM', 'A/AN',
	     'A/AP', 'A/AR', 'A/AS', 'A/AT', 'A/AY',
	     
	     'B', 'B/B0', 'B/B1', 'B/B2', 'B/B3', 'B/B7', 'B/BL', 'B/BM', 'B/BN', 'B/BP',

	     'C', 'C/C8', 'C/C9', 'C/CA', 'C/CB', 'C/CC', 'C/CD', 'C/CE', 'C/CF', 'C/CG', 'C/CH',
	     'C/CJ', 'C/CK', 'C/CL', 'C/CM', 'C/CN', 'C/CP', 'C/CR', 'C/CS', 'C/CT', 'C/CU',

	     'D', 'D/D8', 'D/D9', 'D/DC',

	     'E', 'E/E9', 'E/EA', 'E/EB', 'E/EC', 'E/ED', 'E/EE', 'E/EF', 'E/EG', 'E/EU', 
	     'E/EV', 'E/EW', 'E/EX',

	     'F', 'F/F7', 'F/F8', 'F/F9', 'F/FA', 'F/FB', 'F/FC', 'F/FD', 'F/FE', 'F/FF', 'F/FL',
	     'F/FM', 'F/FN', 'F/FP', 'F/FR', 'F/FS', 'F/FT', 'F/FU', 'F/FX', 'F/FY',

	     'G', 'G/G0', 'G/G1', 'G/G2', 'G/G3', 'G/G4', 'G/G5', 'G/G6', 'G/GS', 'G/GT', 'G/GU',
	     'G/GV', 'G/GW', 'G/GX', 'G/GY',

	     'H', 'H/H0', 'H/H1', 'H/H4', 'H/H5', 'H/H6', 'H/H7', 'H/H8', 'H/H9', 'H/HA', 'H/HB',
	     'H/HC', 'H/HD', 'H/HE', 'H/HF', 'H/HG', 'H/HH', 'H/HJ', 'H/HK', 'H/HL', 'H/HM', 'H/HN',
	     'H/HP', 'H/HR', 'H/HS', 'H/HT', 'H/HU', 'H/HV', 'H/HW', 'H/HX', 'H/HY',

	     'J', 'J/J0', 'J/J1', 'J/J2', 'J/J3', 'J/J4', 'J/J5', 'J/J6', 'J/J7', 'J/J8', 'J/J9',
	     'J/JA', 'J/JJ', 'J/JK', 'J/JN', 'J/JP', 'J/JS', 'J/JT', 'J/JW', 'J/JX', 'J/JY',

	     'K', 'K/K1', 'K/K2', 'K/K3', 'K/K4', 'K/K5', 'K/K6', 'K/K7', 'K/K8', 'K/K9', 'K/KA',
	     'K/KB', 'K/KC', 'K/KD', 'K/KE', 'K/KG', 'K/KJ', 'K/KL', 'K/KM', 'K/KN', 'K/KP', 'K/KR',
	     'K/KS'

	     );


# ensure package returns 1 to keep Perl happy
1;
