#!/usr/bin/perl
#
# SCoRE - Search a Corpus for Regular Expressions
#
# Perl module for CHILDES-specific routines
#
# (c) Matthew Purver, King's College London, 2002
#
# $Log: $
#

package CHILDES;

# export subroutines
require Exporter;
our @ISA = qw( Exporter );
our @EXPORT = qw( checkHeaderType 
		  getHeaderInfo
		  getSentence
		  removeSentence
		  noTags 
		  noTagsAtAll
		  highlight
		  fullPath
		  skipHeader 
		  skipAlign 
		  stringToRegExp );
our @EXPORT_OK = qw( %includeTypes $defaultIncludeType %excludeTypes $defaultExcludeType @dataDirs
		     $corpusFilePath $corpusListPath );

# imports
use SCoRE qw( $corpus $corpusPath $startMark $endMark $startTag $endTag $sentNum $useTurns $startDir $showComments );

# defines
$corpusFilePath = $corpusPath;				# path to corpus files
$corpusListPath = $corpusPath;				# path to corpus file lists
( $corpus eq 'CHILDES' ) and ( $startDir = ( $startDir eq '' ) ? '*/*.cha */*/*.cha' : "$startDir/*.cha" ); # dir spec

# the CHILDES format does not differentiate between turns and sentences, so we convert when reading in ...
$turnChar = '*';
$rawTurnSep = '\*[a-zA-Z]{3}:\s*';		# original turn separator (also used for sents) without numbering
$rawTurnSepC = '\*([a-zA-Z]{3}:\s*)';		# original turn separator (also used for sents) without numbering
$turnSep = '\*\d*\*[a-zA-Z]{3}:\s*';		# turn separator
$turnSepC = '\*(\d*)\*([a-zA-Z]{3}):\s*';	# turn separator, capturing sentNum and speaker ID
$sentSep = '\*\d*\*[a-zA-Z#]{3}:\s*';		# sentence separator (pattern)
$sentSepC = '\*(\d*)\*[a-zA-Z#]{3}:\s*';	# sentence separator (pattern), capturing sentNum
$sentSepA = '"*$1*###: "';			# sentence separator (actual, for insertion)
$uttSep = $useTurns ? $turnSep : $sentSep;	# sentence/turn separator


# discard to end of header section (or fail)
sub skipHeader {
    $lastSpeaker = '';
    $iSent = 0;
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*($rawTurnSep)/$1/ );
    }
    return 0;
}


# dummy function (no align section)
sub skipAlign {
    return 1;
}


# initialize & read to end of header section (or fail)
sub getHeader {
    $lastSpeaker = '';
    $iSent = 0;
    $lineBuffer = '';
    while ( <STDIN> ) {
	chomp( $lineBuffer .= $_ );
	return 1 if ( $lineBuffer =~ /$rawTurnSep/ );
    }
    return 0;
}


# get file description from header
sub getHeaderDescription {
    my $info = '';

    # get any info
    ( $lineBuffer =~ /\@Comment:\s*([^\@\*]+)/ ) and ( $info .= " - $1" );
    ( $lineBuffer =~ /\@Situation:\s*([^\@\*]+)/ ) and ( $info .= " - $1" );
    ( $lineBuffer =~ /\@Activities:\s*([^\@\*]+)/ ) and ( $info .= " - $1" );

    # and return it
    return ( ( $info or 'none' ), 'none' );
}

# no typing in header, so just get file description
sub checkHeaderType {
    # get whole header and description
    getHeader() or return 0;
    my $info = getHeaderDescription();

    # discard header from buffer and return info
    $lineBuffer =~ s/^.*($rawTurnSep)/$turnChar . ++$iSent . $1/e;
    return ( $info, 'none' );
}


# get file description, and set up CP names & corresponding display colours
sub getHeaderInfo {
    my ( $iAnon, $iColour, $key, $value, @cpinfo );

    @colours = ( 'Gray', 'Red', 'Purple', 'Maroon', 'Fuchsia', 
		 'Green', 'Lime', 'Olive', 'Yellow', 'Navy', 'Blue', 'Teal', 'Aqua' );

    # get whole header
    getHeader() or return 0;
    my $info = getHeaderDescription();

    # get CP names from <person> info if present, distinguishing between 'None' entries
    $iAnon = 0;
    $iColour = 0;
    $iCP = -1;
#    %people = ( $lineBuffer =~ /\@ID:[^\@]*?\|([A-Z]{3})\|[^\@]*?\|(\w+)\|/g );
    $lineBuffer =~ /\@Participants:(.*?)(\@|\*)/;
    %people = ( $1 =~ /\b([A-Z][A-Za-z]{2})\b\s*([^,]+)/g );
    while (($key, $value) = each %people) {
	( $cpinfo[++$iCP] = $value ) =~ s/\s+/, /g;
	$cpinfo[$iCP] =~ s/,\s*$//g;
	$people{$key} =~ s/(\w\s+)target_child/$1(Ch)/gi;
	$people{$key} =~ s/(\w\s+)investigator/$1(In)/gi;
#	$people{$key} =~ s/(\w\s+)mother/$1(Mo)/gi;
#	$people{$key} =~ s/(\w\s+)father/$1(Fa)/gi;
	$people{$key} =~ s/(\w\s+)unidentified/$1(?)/gi;
	$people{$key} =~ s/(\w\s+)(\w+)/$1($2)/gi;
	$people{$key} =~ s/\s+/ /g;
	$whoColours{$key} = $colours[($iColour++) % @colours];
	# deal with files where definition = Ang, but ANG used as uttSep
	if ( ( $tmp = $key ) =~ tr/a-z/A-Z/ ) {
	    $people{$tmp} = $people{$key};
	    $whoColours{$tmp} = $whoColours{$key};
	}
    }

    # set up names & colours for global IDs
    $people{ 'par' } = 'Parents';
    $people{ 'com' } = 'Comment';
    $whoColours{ 'par' } = $whoColours{ 'com' } = 'White';

    # discard header from buffer and return info
    $lineBuffer =~ s/^.*($rawTurnSep)/$turnChar . ++$iSent . $1/e;
    return ( $info, @cpinfo );
}


sub noTagsAtAll {
    my @string = @_;

    for ( @string ) {
	s/\@[^\@]*\@//g;
	s/\@[a-zA-Z]*//g;
	s/\[[^\]]*\]//g;
	s/(\s)\#+/$1/g;
	s/\+[\/+]?[,.?!]+//g;
	s/[hxyw]{2,}//g;
	s/$turnSep/| /go;
	s/$sentSep/ /go;
    }

    return @string;
}


# process a sentence SGML -> HTML suitable for display
sub noTags {
    my ( $notags ) = @_;
    my ( $l, $r ) = ( '__LEFT__', '__RIGHT__' );
    my ( $l1, $r1 ) = ( '<em>&lt;', '&gt;</em>' );
    my ( $li, $ri ) = ( '__LEFTI__', '__RIGHTI__' );
    my ( $li1, $ri1 ) = ( '<i>', '</i>' );
    my ( $sp ) = '(?:\s*(?:\[[^\[\]]*\])*)*';
    my $odd = 1;

    for ( $notags ) {

	# remove code tags (marking foreign words)
	s/(\w+)\@s\b/$li$1$ri/g;
	s/(\w)\@[a-zA-Z]*/$1/g;

	# colour-code overlapping speech
	s/<([^<>]*?)>$sp\[>(\d+)?\]/"<font COLOR=" . $colours[newColour($2) % @colours] . ">$1<\/font>"/ge;
	s/(\S+)$sp\[>(\d+)?\]/"<font COLOR=" . $colours[newColour($2) % @colours] . ">$1<\/font>"/ge;
	s/<([^<>]*?)>$sp\[<>(\d+)?\]/"<font COLOR=" . $colours[oldColour($2) % @colours] . ">$1<\/font>"/ge;
	s/(\S+)$sp\[<>(\d+)?\]/"<font COLOR=" . $colours[oldColour($2) % @colours] . ">$1<\/font>"/ge;
	s/<([^<>]*?)>$sp\[<(\d+)?\]/"<font COLOR=" . $colours[oldColour($2) % @colours] . ">$1<\/font>"/ge;
	s/(\S+)$sp\[<(\d+)?\]/"<font COLOR=" . $colours[oldColour($2) % @colours] . ">$1<\/font>"/ge;

	# translate special stuff
	s/\b[0O]\s+\[\=\!\s*([^\[\]]*)\]/$l$1$r/go; # actions
	s/\b[0O]\s+\[:=\w+\s+([^\[\]]*)\]/${l}trans: $1$r/go; # actions
	s/\[\=\!\s*([^\[\]]*)\]/$l$1$r/go; # paralinguistic material
	s/<([^<>]*?)>\s*\[\!+\]/\[${l}stressed$r $1\]/go; # stressing
	s/\b(\S+)\s*\[\!+\]/\[${l}stressed$r $1\]/go; # stressing
	s/\&(\S+)\b/<s>$1<\/s>/g; # phonetic transcriptions
	s/<([^<>]*?)>\s*\[\/[\/\?-]*\]/<s>$1<\/s>/g; # retracing
	s/(\S+)\s*\[\/[\/\?-]*\]/<s>$1<\/s>/g; # retracing
	s/<([^<>]*?)>\s*\[\"\]/\"$1\"/g; # quotation mark (metalinguistic reference)
#	s/(\s)\#+(?=\s)/$1${l}pause$r/go; # pauses
	s/(\s)\#+/$1${l}pause$r/go; # pauses
	s/\+([.?!]+)/$1/g; # trailing off
	s/\+\/+[.?]/--/g; # interruptions
	s/\+,/--/g; # interruptions
	s/<([^<>]*?)>(\s*\[)/$1$2/g; # anything else I've missed?
	s/hh+/${l}breath$r/go; # in/out-breath
	s/xx+/${l}unclear$r/go; # unclear
	s/yy+/${l}unclear$r/go; # phonetically transcribed on %phon tier
	s/ww+/${l}untranscribed$r/go; # untranscribed
        s/\[:=\w+\s+([^\[\]]+)\]/${l}trans: $1$r/go;

	# show comments
	s/\[comment value=\"(.*?)\"\]/$showComments ? " <em>[$1]<\/em>" : ''/ge;

	# replace <div>, <bncDoc> boundaries
	s/\@Time Start:\s+([^\@]*)/<\/td><\/tr><tr><td colspan=3>----------------- $1 -----------------/o;
	s/\@Location:\s+([^\@]*)/<\/td><\/tr><tr><td colspan=3>----------------- $1 -----------------/o;
	s/\@Activities:\s+([^\@]*)/<\/td><\/tr><tr><td colspan=3>----------------- $1 -----------------/o;
	s/\@Situation:\s+([^\@]*)/<\/td><\/tr><tr><td colspan=3>----------------- $1 -----------------/o;
	s/\@New(\s+)?Episode/<\/td><\/tr><tr><td colspan=3>----------------- New Conversation -----------------/o;
	s/\@End(.*)/<\/td><\/tr><tr><td colspan=3>----------------- End$1 -----------------/o;

#  	# make tags visible
#  	s/<(.+?)>/$l$1$r/go;

	# table formatting for turn/sent tags
	s/$turnSepC/"<\/td><\/tr><tr><td nowrap bgcolor=$whoColours{$2}><em>$people{$2}<\/em><\/td><td>" . ( $sentNum ? "<font size=-2>$1<\/font>" : '' ) . "<\/td><td>"/goe;
	s/$sentSepC/"<\/td><\/tr><tr><td><\/td><td>" . ( $sentNum ? "<font size=-2>$1<\/font>" : '' ) . "<\/td><td>"/goe;

	# finally, replace __B/E__ markers
	s/$startMark/$startTag/go;
	s/$endMark/$endTag/go;
	s/$l/$l1/go;
	s/$r/$r1/go;
	s/$li/$li1/go;
	s/$ri/$ri1/go;

    }

    for ( $notags ) {
    }


    return $notags;
}


$ptrColour = 0;

sub newColour {
    my ( $n ) = @_;
    $ptrColour++;
    defined( $n ) and ( $ptrColour[$n] = $ptrColour );
    return $ptrColour;
}

sub oldColour {
    my ( $n ) = @_;
    return ( defined( $n ) ? $ptrColour[$n] : $ptrColour );
}

# convert search string to reg-exp suitable for searching SGML
sub stringToRegExp {
    my ( $arg, $caseSens, $punctSens, $strictPoS ) = @_;

    $pu = '[,.?!;:]';
    $sp = '[<>]?' . ( $punctSens ? '' : "\\s*$pu?" ) . '\s*(?:\s*\[[^\[\]]*\])*\s*[<>]?';
    $cs = ( $caseSens ? '' : '(?i)' );
    $br = $uttSep;
    $wd = '[\'\w]'; # words may contain ' chars (e.g. n't)
    $tg = '(?:\[[^\[\]]*\]\s*)';
    $at = '(?:@\w{1,3})?'; # special form markers
    my $nSent = 0;

    # translate wildcards to regexp
    for ( $arg ) {

	# first count the number of sentences/turns
	$nSent += ( @tmp = /(^|\s)\|/g );

	# replace whitespace by tabs to make it easier to replace uniquely later
	s/\s+/\t/g;

	# capture all terms
	s/\(/\(\?:/g;         # ( -> (?:  (to prevent grouping also gathering)
	s/(\S+)/($1)/g;

	# escape punctuation mark terms
	s/\(($pu)\)/(\\$1)/go;

	# add boundaries to words
	s/\(([\w\',?*+():|^]+)\)/(\\b$1\\b)/go;
	s/\\b([|^])\\b/$1/go;

	# convert wildcards to regexp equivalents
	s/(?<!\])\*/$wd\*/go;        # * -> ['\w]*
	s/(?<!\])\+/$wd\+/go;        # + -> ['\w]+
	s/([^\\)])\?([^:])/$1$wd$2/go;    # ? -> ['\w]  (unless escaped or old [])
	s/([^\\]),(\\b\))/$1$2\t.*?/go;  # , -> .*? (non-escaped comma last in term)

        # spot PoS tag searches
        /</ and print "Sorry - PoS searches not yet implemented in CHILDES" and die "PoS search attempted";

        # add tags to words (rearranging optional brackets)
        s/\\b\)/\\b)$at/go;
        s/(\(\\b)(\(\?:)([\w\',?*+():|^]+)(\)\?)(\\b\)$at)/$2$1$3$5$4/go;

	# deal with sentence/turn breaks (don't allow .* if ^, $ adjacent)
         s/(\$\)\t)\(\|\)(\t\(\^)/$1($br)$2/go;
         s/(^|\t)\(\|\)(\t\(\^)/$1.*?($br)$2/go;
         s/(\$\t)\(\|\)(\t|$)/$1($br).*?$2/go;
         s/(^|\t)\(\|\)(\t|$)/$1.*?($br).*?$2/go;

#          s/(\(\$\))\t(\(\|\))/$pu?\\s*$tg*$2/go;
#          s/([^\$]\))\t(\(\|\))/$1.*?$2/go;
#          s/(\(\|\))\t(\(\^\))/$1$tg*/go;
#          s/(\(\|\))\t(\([^\^])/$1.*?$2/go;
#          s/\(\|\)/($br)/go;


  	# add required tags to ^, $, keeping (^/$) only if actually at beginning/end
#	s/(^)(\(\^\))(\t|$)/^(\\*\\d+\\*[A-Z]{3}:)\\s*$tg*$3/go; # HACK must not = $uttSep
	s/(^)(\(\^\))(\t|$)/^($uttSep)\\s*$tg*$3/go; # why not use $uttSep?
	s/(\t)(\(\^\))(\t|$)/$1\\s*()$tg*$3/go;
        s/(^|\t)(\(\$\))($)/$1$pu?\\s*$tg*\$$3/go;
        s/(^|\t)(\(\$\))(\t)/$1$pu?\\s*$tg*()$3/go;

#   	s/^\(\^\)/(^)\\s*$br\\s*$tg*/go;
#          s/\(\$\)$/$pu?\\s*$tg*(\$)/go;

	# case sens
	s/(\((\?\:)?)(.*?)\)/$1$cs$3)/go;

	# and replace tabs with required separator
	s/(^|\t)(\.\*\?)(\t|$)/$2/go;
	s/\t/$sp/go;

	# now put brackets round the whole thing for capturing (UNLESS empty!)
	# (remembering to increment all capture-counters ...)
	s/^/(/;
	s/$/)/;
	s/^\(\)$//;
	s/\\(\d+)/'\\' . ($1+1)/ge;

    }
    return ( $arg, $nSent );

}


# convert turn markers to sentence markers where appropriate, keeping track of whose turn it is
sub tidyLine {
    my( $line ) = @_;
    
    # there will only be one turn separator per line, thank goodness
    $line =~ s/(?!^)$rawTurnSepC/$1/go;
    $line =~ s/($rawTurnSep)/$turnChar . ++$iSent . $1/e;
    $line =~ /$turnSepC/o and ( $2 eq $lastSpeaker ) and ( $line =~ s/$turnSepC/$sentSepA/eeo );
    $line =~ /$turnSepC/o and ( $lastSpeaker = $2 );
    return $line;
}


sub getSentence {

    while ( <STDIN> ) {
  	chomp( $lineBuffer .= tidyLine( $_ ) );

	# mark comments properly
	$lineBuffer =~ s/\[%\s*(.*?)\]/[comment value=\"$1\"]/g;
	$lineBuffer =~ s/\@Comment:?\s+(.*?)(?=$sentSep|[%@])/[comment value=\"$1\"]/go;
	while ( $lineBuffer =~ /%(.*?)(?=$sentSep|[%@])/ ) {
	    ( $tmp = $1 ) =~ s/<(.*?)>/($1)/g;
	    $lineBuffer =~ s/%(.*?)(?=$sentSep|[%@])/[comment value=\"$tmp\"]/go;
	}
	$lineBuffer =~ s/\[\+\s*(.*?)\]/[comment value=\"$1\"]/g;

	# beginning of next sentence? return complete one
	if ( $lineBuffer =~ /$uttSep.*$uttSep/o ) {
	    $lineBuffer =~ s/^(.*?$uttSep.*)($uttSep)/$2/o;
	    return $1;
	}
    }
    
    # end of file? return any dregs
    $lineBuffer =~ s/^(.*)$//o;
    return $1;

}


sub removeSentence {
    my( $string ) = @_;

    # remove first sentence (or whole thing) from buffer string
    ( $string =~ s/(^.*?$uttSep.+?)($uttSep)/$2/o ) or ( $string =~ s/(.*)//o );

    # return what's left and what was removed )
    return ( $string, $1 );
}


sub highlight {
    my( $text ) = @_;

    for ( $text ) {
	s/^/$startMark/o;
	s/$/$endMark/o;
	s/($sentSep)/$1$startMark/go;
    }
    return $text;
}


sub fullPath {
    my( $file ) = @_;

    ( $file =~ /$corpusFilePath/ ) or ( $file = "$corpusFilePath/$file" );
    return $file;
}


%includeTypes = ();
$defaultIncludeType = '';
%excludeTypes = ();
$defaultExcludeType = '';
@dataDirs = (
	     'bates/free20',
	     'bates/free28',
	     'bates/snack28',
	     'bates/story28',

	     'belfast',

	     'bernstein/children',
	     'bernstein/intervie',

	     'bliss',

	     'bloom70/eric',
	     'bloom70/peter',

	     'bloom73',

	     'bohannon/bax',
	     'bohannon/nat',

	     'braine',

	     'brown/adam',
	     'brown/eve',
	     'brown/sarah',

	     'carter',

	     'clark',

	     'cornell/haas',
	     'cornell/horn',
	     'cornell/lnso',
	     'cornell/mom',
	     'cornell/moore',
	     'cornell/nurse',
	     'cornell/sch',
	     'cornell/woot',

	     'cruttenden',

	     'demetras1',

	     'demetras2',

	     'evans',

	     'evesyn',

	     'feldman',

	     'fletcher/3',
	     'fletcher/5',
	     'fletcher/7',

	     'garvey',

	     'gathburn',

	     'gathercole',

	     'gleason/dinner',
	     'gleason/father',
	     'gleason/mother',

	     'haggerty',

	     'hall/blackpro',
	     'hall/blackwor',
	     'hall/whitepro',
	     'hall/whitewor',

	     'higginson',

	     'howe',

	     'korman',

	     'kuczaj',

	     'macwhinney/BOYS',
	     'macwhinney/CUTECLIP',
	     'macwhinney/ROSS',
	     'macwhinney/UNFINISH',

	     'manchester/anne',
	     'manchester/aran',
	     'manchester/becky',
	     'manchester/carl',
	     'manchester/domin',
	     'manchester/gail',
	     'manchester/joel',
	     'manchester/john',
	     'manchester/liz',
	     'manchester/nic',
	     'manchester/ruth',
	     'manchester/warr',

	     'morisset/seattle',
	     'morisset/topeka',
	     'morisset/ucla',

	     'nelson',

	     'newengland/ne14',
	     'newengland/ne20',
	     'newengland/ne32',

	     'peters',

	     'post',

	     'sachs',

	     'snow',

	     'suppes',

	     'tardiff',

	     'valian',

	     'vanhouten/threes',
	     'vanhouten/twos',

	     'vankleeck',

	     'warren',

	     'wells'
	     );

# ensure package returns 1 to keep Perl happy
1;
