#!/usr/bin/perl
#
# SCoRE - Search a Corpus for Regular Expressions
#
# Perl module for TRAINS-specific routines
#
# (c) Matthew Purver, King's College London, 2002
#
# $Log: $
#

package TRAINS;

# export subroutines
require Exporter;
our @ISA = qw( Exporter );
our @EXPORT = qw( checkHeaderType 
		  getHeaderInfo
		  getSentence
		  removeSentence
		  noTags 
		  noTagsAtAll
		  highlight
		  fullPath
		  skipHeader 
		  skipAlign 
		  stringToRegExp );
our @EXPORT_OK = qw( %includeTypes $defaultIncludeType %excludeTypes $defaultExcludeType @dataDirs
		     $corpusFilePath $corpusListPath );

# imports
use SCoRE qw( $corpusPath $startTag $endTag $sentNum $useTurns $startDir $showComments );
$startMark = '##B##'; # don't use standard __B__ one, as underscores used in TRAINS
$endMark = '##E##';

# defines
$corpusFilePath = $corpusPath;				# path to corpus files
$corpusListPath = $corpusPath;				# path to corpus file lists
$startDir = ( $startDir eq '' ) ? '*/*' : "$startDir/*"; # modify default dir spec

# 91 & 93 formats have quite different separators ...
$turnSep = '(?:utt\d+\s*:\s*[su]:|\d+\.\d+\s*[SM]:)';		# turn separator
$turnSepC = '(?:utt(\d+)\s*:\s*([su]):|(\d+\.\d+)\s*([SM]):)';	# turn separator, capturing sentNum / ID
$sentSep = '(?:utt\d+\s*:(?:\s*[su]:)?|\d+\.\d+\s*(?:[SM])?:)';	# sentence separator
$sentSepC = '(?:utt(\d+)\s*:(?:\s*[su]:)?|(\d+\.\d+)\s*(?:[SM])?:)';# sentence separator, capturing sentNum
$uttSep = $useTurns ? $turnSep : $sentSep;				# sentence/turn separator


# initialise & discard to end of header section (or fail)
sub skipHeader {
    $lineBuffer = '';
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	$lineBuffer .= ' ';
	return 1 if ( $lineBuffer =~ /$uttSep/ );
    }
    return 0;
}


# dummy function (no align section)
sub skipAlign {
    return 1;
}


# no interesting header info available, so just skip
sub checkHeaderType {
    skipHeader() or return 0;
    return ( 'none', 'none' );
}


# set up CP names & corresponding display colours
sub getHeaderInfo {

#      @colours = ( 'Gray', 'Red', 'Purple', 'Maroon', 'Fuchsia', 
#  		 'Green', 'Lime', 'Olive', 'Yellow', 'Navy', 'Blue', 'Teal', 'Aqua' );

    # skip header
    skipHeader() or return 0;

    # set up names & colours for global IDs
    # first for '91 dialogues
    $people{ 'S' } = 'System';
    $people{ 'M' } = 'User';
    $whoColours{ 'S' } = 'Blue';
    $whoColours{ 'M' } = 'Green';
    # now for '93 dialogues
    $people{ 's' } = 'System';
    $people{ 'u' } = 'User';
    $whoColours{ 's' } = 'Blue';
    $whoColours{ 'u' } = 'Green';
    my @cpinfo = ( 'System', 'User' );

    # discard header from buffer and return (empty) info
    $lineBuffer =~ s/^.*($uttSep)/$1/;
    return ( 'none', 'none', @cpinfo );
}


sub noTagsAtAll {
    my @string = @_;

    for ( @string ) {
	s/<[^>]*?>//g;
	s/\[[^\]]*?\]//g;
	s/\+[^\+]*?\+//g;
	s/_[^_]*?_//g;
	s/$turnSep/| /go;
	s/$sentSep/ /go;
    }

    return @string;
}


# process a sentence SGML -> HTML suitable for display
sub noTags {
    my ( $notags ) = @_;
    my $l = '<em>&lt;';   # MUST contain > chars
    my $r = '&gt;</em>';  # MUST contain > chars
    my $odd = 1;
    my $lastNum = 1;

    ( $ptrColour = 0 ) unless defined $ptrColour;

    for ( $notags ) {
	# make <>, [] tags visible
	s/<([^>]*?)>/$l$1$r/g;
	s/\[([^\]]*?)\]/$l$1$r/g;

	# change +, _ tags
	s/\+([^\+]*?)\+/<em>($1)<\/em>/g;
	s/_([^_]*?)_/<em>[$1]<\/em>/g;

	# replace ##B/E## markers
	s/$startMark/$startTag/go;
	s/$endMark/$endTag/go;
    }

    # table formatting for utterances/turns
    for ( $notags ) {

	s/$turnSepC/"<\/td><\/tr>\n<tr><td bgcolor=" . $whoColours{($2 or $4)} . '><em>' . $people{($2 or $4)} . '<\/em><\/td><td>' . ( $sentNum ? ( '<font size=-2>' . ( $1 or $3 ) . '<\/font>' ) : '' ) . '<\/td><td>'/goe;
	s/$sentSepC/"<\/td><\/tr>\n<tr><td><\/td><td>" . ( $sentNum ? ( '<font size=-2>' . ( $1 or $2 ) . '<\/font>' ) : '' ) . '<\/td><td>'/goe;

    }

    return $notags;
}


# convert search string to reg-exp suitable for searching SGML
sub stringToRegExp {
    my ( $arg, $caseSens, $punctSens, $strictPoS ) = @_;

    $pu = '[,.?!;:]';
    $sp = ( $punctSens ? '\s*' : "$pu?\\s*" );
    $cs = ( $caseSens ? '' : '(?i)' );
    $br = $uttSep;
    $wd = '[\'\w]'; # words may contain ' chars (e.g. n't)
    $mk = '[+_]?';   # optional marker chars
    $mk1 = '\[\+_\]\?';   # optional marker chars
    my $nSent = 0;

    # translate wildcards to regexp
    for ( $arg ) {

	# first count the number of sentences/turns
	$nSent += ( @tmp = /(^|\s)\|/g );

	# replace whitespace by tabs to make it easier to replace uniquely later
	s/\s+/\t/go;

	# convert wildcards to regexp equivalents
	s/\(/\(\?:/g;         # ( -> (?:  (to prevent grouping also gathering)
	s/(?<!\])\*/$wd\*/go;        # * -> ['\w]*
	s/(?<!\])\+/$wd\+/go;        # + -> ['\w]+
	s/([^\s\(\)])\?/$1$wd/go;    # ? -> ['\w]  (unless preceded by whitespace/bracket)
	s/([^\(\)])\?(\S)/$1$wd$2/go;    # ? -> ['\w]  (unless preceded by bracket/followed by whitespace)
	s/(\S+),/$1\t\.\*\?/go;  # , -> .*? (if not preceded by whitespace)

        # spot PoS tag searches
        /</ and print "Sorry - no PoS tagging in TRAINS" and die "PoS search attempted";

	# deal with sentence/turn breaks (don't allow .* if ^, $ adjacent)
        s/(\$\t)(\|)(\t\^)/$1($br)$3/go;
        s/(^|\t)(\|)(\t\^)/$1(.*?$br)$3/go;
        s/(\$\t)(\|)(\t|$)/$1($br.*?)$3/go;
        s/(^|\t)(\|)(\t|$)/$1(.*?$br.*?)$3/go;

	# add required tags to ^, $, keeping (^/$) only if actually at beginning/end
	s/(^)(\^)(\t|$)/($2)\\s*(?:<[^<>]*>\\s*)*$3/go;
	s/(\t)(\^)(\t|$)/$1\\s*(?:<[^<>]*>\\s*)*$3/go;
        s/(^|\t)(\$)($)/$1$pu?\\s*(?:<[^<>]*>\\s*)*($2)$3/go;
        s/(^|\t)(\$)(\t)/$1$pu?\\s*(?:<[^<>]*>\\s*)*$3/go;

	# mark punctuation marks to prevent \b
	s/(^|\t)($pu)(\t|$)/$1<$2>/go;

        # add boundaries to words & capture (rearranging optional brackets)
  	s/(^|\t)([^<\t|]+)(?=\t|$)/$1($cs$mk\\b$mk$2$mk\\b$mk)/go;
  	s/($mk1\\b$mk1)(\(\?:)([^<\t|]+)(\)\?)($mk1\\b$mk1)/$2$1$3$5$4/go;

	# escape punctuation marks & capture
	s/<($pu)>/(\\$1)/go;

	# and replace tabs with required separator
	s/(^|\t)(\.\*\?)(\t|$)/$2/go;
	s/\t/$sp/go;

	# now put brackets round the whole thing for capturing (UNLESS empty)
	# (remembering to increment all capture-counters ...)
	s/^/(/;
	s/$/)/;
	s/^\(\)$//;
	s/\\(\d+)/'\\' . ($1+1)/ge;

    }
    return ( $arg, $nSent );

}


sub getSentence {

    while ( <STDIN> ) {
  	chomp( $lineBuffer .= $_ );
	$lineBuffer .= ' '; # add line-break space (as we don't have decent utterance separator)
	if ( $showComments ) {
	    $lineBuffer =~ s/\%+(.*)$/<comment>$1<\/comment>/; # mark comments properly
	}
	else {
	    $lineBuffer =~ s/\%.*$//; # remove comments
	}

    	# beginning of next sentence? return complete one
  	if ( $lineBuffer =~ /($uttSep).*?($uttSep)/o ) {
  	    $lineBuffer =~ s/^(.*?($uttSep).*?)($uttSep)/$3/o;
  	    return $1;
  	}
    }
    
    # end of file? return any dregs
    $lineBuffer =~ s/^(.*)$//o;
    return $1;

}


sub removeSentence {
    my( $string ) = @_;

    # remove first sentence (or whole thing) from buffer string
    ( $string =~ s/(^.*?$uttSep.+?)($uttSep)/$2/o ) or ( $string =~ s/(.*)//o );

    # return what's left and what was removed )
    return ( $string, $1 );
}


sub highlight {
    my( $text ) = @_;

    for ( $text ) {
	s/^/$startMark/o;
	s/$/$endMark/o;
	s/($sentSep)/$1$startMark/go;
    }
    return $text;
}


sub fullPath {
    my( $file ) = @_;

    ( $file =~ /$corpusFilePath/ ) or ( $file = "$corpusFilePath/$file" );
    return $file;
}


%includeTypes = ();
$defaultIncludeType = '';
%excludeTypes = ();
$defaultExcludeType = '';
@dataDirs = ( '91', '93' );

# ensure package returns 1 to keep Perl happy
1;
